
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>


enum {
//	_xinfreq = 20_000_000,
	_xtlfreq = 20_000_000,
	_clkfreq = 200_000_000,

//
// Prop2 boot pins, hardcoded in the Prop2 ROM
//
    MISO_EVAL = 58, MOSI_EVAL, CS_EVAL, CLK_EVAL,

//
// Roger Loh's 4-bit add-on board
//
    BASE_PIN_RL = 16,
    DAT0_RL     = BASE_PIN_RL+0,    // MISO, DO
    DAT1_RL     = BASE_PIN_RL+1,
    DAT2_RL     = BASE_PIN_RL+2,
    DAT3_RL     = BASE_PIN_RL+3,    // CS
    CMD_RL      = BASE_PIN_RL+4,    // MOSI, DI
    CLK_RL      = BASE_PIN_RL+5,    // SCLK
    LED_RL      = BASE_PIN_RL+6,    // output to red LED
    PWR_RL      = BASE_PIN_RL+7,    // Card-Detect input, and power switch, and green LED

    CS_RL       = DAT3_RL,
    MOSI_RL     = CMD_RL,
    MISO_RL     = DAT0_RL,

//
// My hand wired full sized 4-bit SD slot
//
    BASE_PIN_EH = 40,
    LED_EH      = BASE_PIN_EH+1,    // output to red LED
    CMD_EH      = BASE_PIN_EH+2,    // MOSI, DI
    CLK_EH      = BASE_PIN_EH+3,    // SCLK
    DAT0_EH     = BASE_PIN_EH+4,    // MISO, DO
    DAT1_EH     = BASE_PIN_EH+5,
    DAT2_EH     = BASE_PIN_EH+6,
    DAT3_EH     = BASE_PIN_EH+7,    // CS

    CS_EH       = DAT3_EH,
    MOSI_EH     = CMD_EH,
    MISO_EH     = DAT0_EH,


    BUFSIZE = 1 * 512,
    FILES = 200,    // Max of 1000
};


int32_t  xoro32state;



static void  randfill( uint32_t *addr, size_t size )
{
    int32_t  state = xoro32state;
    size >>= 2;

    __asm volatile {    // "volatile" enforces unoptimised use of FCACHE
		wrfast	#0, addr
		rep	@.rend, size
		xoro32	state
		mov	pb, 0-0
		wflong	pb
.rend
    }
    xoro32state = state;
}


static size_t  randcmp( const uint32_t *addr, size_t size )
{
    int32_t  state = xoro32state;
    size_t  pass = size >> 2;

    __asm volatile {    // "volatile" enforces unoptimised use of FCACHE
		rdfast	#0, addr
.loop
		rflong	pb
		xoro32	state
		cmp	pb, 0-0  wz
	if_z	djnz	pass, #.loop
    }
    xoro32state = state;

    return  size - (pass << 2);    // matched amount
}


int  nullcheck( void *buf, size_t longs )
{
    __asm volatile {    // "volatile" enforces unoptimised use of FCACHE
		rdfast	#0, buf
		modcz	0xff   wz
		rep	@.rend, longs
	if_z	rflong	pb   wz
.rend
		wrz	longs
    }
    return longs;
}




static void  writefile( const char *filename, uint32_t *data )
{
    FILE  *fh;
    int  rc;

    errno = 0;
    fh = fopen(filename, "wb");
    if( !fh ) {
        perror("fopen() for writing failed");
        exit(1);
    }

    randfill(data, BUFSIZE);

    rc = fwrite(data, 1, BUFSIZE, fh);
    if( !rc ) {
        perror("fwrite() failed");
        exit(1);
    }

    fclose(fh);

}



static void  checkfile( const char *filename, uint32_t *data )
{
    FILE  *fh;
    int  rc;

    memset(data, 0, BUFSIZE);  // erase any echoes

    errno = 0;
    fh = fopen(filename, "rb");
    if( !fh ) {
        perror("fopen() for reading failed");
        exit(1);
    }

    rc = fread(data, 1, BUFSIZE, fh);
    if( !rc ) {
        perror("fread() failed");
        exit(1);
    }
    fclose(fh);

    if( BUFSIZE != randcmp(data, BUFSIZE) ) {
        perror("rancmp() failed");
        exit(2);
    }
}



static void  rmfile( const char *filename, uint32_t *data )
{
    remove(filename);
}



//===================================================================================

char filepath1[] = "/sd1/copy000.bin";


static void  oldtester( void )
{
    uint32_t  *data = __builtin_alloca(BUFSIZE);    // auto-frees upon return
    uint32_t  rate, ticks, ticke, msecs, onesecond, seed;
    int  i;
//    char  numstr[8];

    onesecond = _clockfreq();
    seed = _rnd();
    xoro32state = seed;

    msecs = 0;
    ticks = _cnt();
    for( i = 0; i < FILES; i++ )  {
        filepath1[9] = i/100 + '0';
        filepath1[10] = i/10%10 + '0';
        filepath1[11] = i%10 + '0';
//        snprintf(numstr, 4, "%03d", i);
//        strncpy(&filepath1[9], numstr, 3);
//        puts(filepath1);
//        exit(0);

        _txraw('.');
        seed = xoro32state;
        remove(filepath1);
        writefile(filepath1, data);
        xoro32state = seed;
        checkfile(filepath1, data);

        ticke = _cnt();
        if(ticke - ticks >= onesecond)  {    // one second chunk
            ticks += onesecond;    // one second in sysclock ticks
            msecs += 1000;    // one second in milliseconds
        }
    }
    msecs += _muldiv64(ticke - ticks, 1000, onesecond);  // remaining ticks to milliseconds
    rate = _muldiv64(FILES, 100000, msecs);
    printf("\n Duration = %d, %d.%02d files/s\n", msecs, rate/100, rate%100);

}



static void  tester(
        void (*callback)(const char *filename, uint32_t *data),
        uint32_t *data )
{
    uint32_t  rate, ticks, ticke, msecs, onesecond;
    int  i;

    onesecond = _clockfreq();
    msecs = 0;
    ticks = _cnt();
    for( i = 0; i < FILES; i++ )  {
        filepath1[9] = i/100 + '0';
        filepath1[10] = i/10%10 + '0';
        filepath1[11] = i%10 + '0';

        callback(filepath1, data);

        ticke = _cnt();
        if(ticke - ticks >= onesecond)  {    // one second chunk
            ticks += onesecond;    // one second in sysclock ticks
            msecs += 1000;    // one second in milliseconds
        }
    }
    msecs += _muldiv64(ticke - ticks, 1000, onesecond);  // remaining ticks to milliseconds
    rate = _muldiv64(FILES, 10000, msecs);
    printf("Duration %d ms, %d.%d files/s\n", msecs, rate/10, rate%10);
}

//===================================================================================


static void  mountall( void )
{
//    typedef struct __using("blkdrvr/sdmm.cc") drv_t;
//    typedef struct __using("blkdrvr/sdmm_bashed.cc") drv_t;
    typedef struct __using("blkdrvr/sdsd.cc") drv_t;
//    typedef struct __using("blkdrvr/sdsd_1.2.cc") drv_t;
    FILE *handle;
    int  clkdiv;

    umount("/sd1");
    drv_t *DRV1 = _gc_alloc_managed(sizeof(drv_t));

    _seterror(0);
//    handle = DRV1->_sdmm_open(CLK_EVAL, CS_EVAL, MOSI_EVAL, MISO_EVAL);
//    handle = DRV1->_sdmm_open(CLK_RL, CS_RL, MOSI_RL, MISO_RL);
    handle = DRV1->_sdsd_open(CLK_RL, CMD_RL, DAT0_RL, PWR_RL, LED_RL);
//    handle = DRV1->_sdsd_open(CLK_EH, CMD_EH, DAT0_EH, -1, LED_EH);
    if( !handle )  {
        printf(" /sd1 open failed!   errno = %d: %s\n", errno, strerror(errno));
        exit(1);
    }
    printf(" /sd1 handle = %x ... ", handle);
    clkdiv = 0;
//    _ioctl(handle, 70, &clkdiv);    // disable read-block CRC processing
    clkdiv = 3;
//    _ioctl(handle, 72, &clkdiv);    // set CLK_DIV value
    mount("/sd1", _vfs_open_fat_handle(handle));
    puts("mounted\n");
}


static void  shutdown( void )
{
    printf("\nClean-up ... errno = %d: %s\n", errno, strerror(errno));

    umount("/sd1");

    _waitms(100);
    _clkset(1, 20_000);  // cool running
}


void  main( void )
{
    uint32_t  seed, *data = __builtin_alloca(BUFSIZE);    // auto-frees upon return

    printf("\n   clkfreq = %d   clkmode = 0x%x\n\n", _clockfreq(), _clockmode());
    atexit(shutdown);

    mountall();

    printf(" Delete %d files ... ", FILES);
    tester(&rmfile, data);

    printf("   File size = %d bytes\n", BUFSIZE);
    printf(" Create %d files ... ", FILES);
    seed = _rnd();
    xoro32state = seed;
    tester(&writefile, data);

    printf(" Verify %d files ... ", FILES);
    xoro32state = seed;
    tester(&checkfile, data);

    exit(0);
}
